from ._spatial_communication import spatial_communication
from ._spatial_communication import communication_direction
from ._spatial_communication import cluster_communication
from ._spatial_communication import cluster_communication_spatial_permutation
from ._spatial_communication import cluster_position
from ._downstream_analysis import communication_deg_detection
from ._downstream_analysis import communication_deg_clustering
from ._downstream_analysis import communication_impact
from ._downstream_analysis import group_cluster_communication
from ._downstream_analysis import group_cell_communication
from ._downstream_analysis import group_communication_direction
from ._downstream_analysis import communication_spatial_autocorrelation